import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver


def compute_regmean_2D_cpu(var, weights):
    weighted_var, vector_weight = variable_weight_2D(var, weights)
    reduced_weight = sum_red.reduce(vector_weight, axis=0)
    reduced_weighted_var = sum_red.reduce(weighted_var, axis=1)
    regmean = regmean_cpu(reduced_weighted_var, reduced_weight)
    return regmean


def compute_regmean_3D_cpu(var, weights):
    weighted_var, vector_weight = variable_weight_3D(var, weights)
    reduced_weight = sum_red.reduce(vector_weight, axis=0)
    reduced_weighted_var = sum_red.reduce(weighted_var, axis=1)
    regmean = regmean_cpu(reduced_weighted_var, reduced_weight)
    return regmean


def compute_regmean_2D_gpu(var, weights):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_weights = cuda.to_device(weights.astype(np.float32))
    gpu_weighted_var = cuda.device_array((times, lats*lons), dtype=np.float32)
    gpu_vector_weight = cuda.device_array(lats*lons, dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_weights, gpu_weighted_var,
                                      gpu_vector_weight)
    del gpu_var, gpu_weights

    reduced_weight = sum_red_cuda(gpu_vector_weight)
    del gpu_vector_weight

    regmean = np.empty(times, dtype=np.float32)
    for time in range(times):
        reduced_weighted_var = sum_red_cuda(gpu_weighted_var[time, :])
        regmean[time] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var

    return regmean


def compute_regmean_gpu_3D(var, weights):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, levs)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_weights = cuda.to_device(weights.astype(np.float32))

    gpu_weighted_var = cuda.device_array((times, levs*lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array(levs*lats*lons, dtype=np.float32)

    variable_weight_cuda_3D[grid, block](gpu_var, gpu_weights,
                                         gpu_weighted_var, gpu_vector_weight)
    del gpu_var, gpu_weights

    reduced_weight = sum_red_cuda(gpu_vector_weight)
    del gpu_vector_weight

    regmean = np.empty(times, dtype=np.float32)
    for time in range(times):
        reduced_weighted_var = sum_red_cuda(gpu_weighted_var[time, :])
        regmean[time] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var

    return regmean


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_weights_2D(mask, e1t, e2t):
    weights = mask*e1t*e2t
    return weights


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weights_3D(mask, e1t, e2t, e3t):
    weights = mask*e1t*e2t*e3t
    return weights


@njit
def variable_weight_2D(var, weight):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    weighted_var = np.empty((times, lats*lons))
    vector_weight = np.empty(lats*lons)
    for t in range(times):
        for j in range(lats):
            for i in range(lons):
                x = i+j*var.shape[2]
                weight_value = weight[0, j, i]
                value = var[t, j, i] * weight_value
                vector_weight[x] = weight_value
                weighted_var[t, x] = value
    return weighted_var, vector_weight


@njit
def variable_weight_3D(var, weight):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    weighted_var = np.empty((times, levs*lats*lons))
    vector_weight = np.empty(levs*lats*lons)
    for t in range(times):
        for l in range(levs):
            for j in range(lats):
                for i in range(lons):
                    x = i+lons*(j+lats*l)
                    weight_value = weight[0, l, j, i]
                    value = var[t, l, j, i] * weight_value
                    vector_weight[x] = weight_value
                    weighted_var[t, x] = value
    return weighted_var, vector_weight


@njit
def variable_weight_level_3D(var, weight):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    weighted_var = np.empty((times, levs, lats*lons))
    vector_weight = np.empty((levs, lats*lons))
    for t in range(times):
        for l in range(levs):
            for j in range(lats):
                for i in range(lons):
                    x = i+j*lons
                    weight_value = weight[0, l, j, i]
                    value = var[t, l, j, i] * weight_value
                    vector_weight[l, x] = weight_value
                    weighted_var[t, l, x] = value
    return weighted_var, vector_weight


@cuda.jit
def variable_weight_cuda(var, weight, weighted_var, vector_weight):
    i, j, t = cuda.grid(3)
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    vector_weight[x] = 0.0
    weighted_var[t, x] = 0.0
    vector_weight[x] = weight[0, j, i]
    weighted_var[t, x] = var[t, j, i] * weight[0, j, i]


@cuda.jit
def variable_weight_cuda_3D(var, weight, weighted_var, vector_weight):
    i, j, lev = cuda.grid(3)
    x = i+var.shape[3]*(j+var.shape[2]*lev)
    if(i >= var.shape[3]):
        return
    vector_weight[x] = 0.0
    vector_weight[x] = weight[0, lev, j, i]
    for t in var.shape[0]:
        weighted_var[t, x] = 0.0
        weighted_var[t, x] = var[t, lev, j, i] * weight[0, lev, j, i]


@cuda.jit
def variable_weight_cuda_level_3D(var, weight, weighted_var, vector_weight):
    i, j, lev = cuda.grid(3)
    x = i+j*var.shape[3]
    if(i >= var.shape[3]):
        return
    for t in range(var.shape[0]):
        vector_weight[lev, x] = 0.0
        vector_weight[lev, x] = weight[0, lev, j, i]
        weighted_var[t, lev, x] = 0.0
        weighted_var[t, lev, x] = var[t, lev, j, i] * weight[0, lev, j, i]


@vectorize(['float64(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def sum_red(x, y):
    return x+y


@vectorize(['float32(float64, float64)'], target='cpu')
def regmean_cpu(weighted_var, weights):
    regmean = weighted_var/weights
    return regmean


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y
